# -*- coding: utf-8 -*-

import pytest
from pmdarima.preprocessing import base
from pmdarima.compat.pytest import pytest_error_str


def test_value_error_on_update_check():
    with pytest.raises(ValueError) as ve:
        base.UpdatableMixin()._check_endog(None)
    assert 'cannot be None' in pytest_error_str(ve)



# -*- coding: utf-8 -*-

import pmdarima as pm
from pmdarima.utils._show_versions import _get_deps_info


# Just show it doesn't blow up...
def test_show_versions():
    pm.show_versions()


def test_show_versions_when_not_present():
    deps = ['big-ol-fake-pkg']
    assert _get_deps_info(deps=deps)['big-ol-fake-pkg'] is None



# -*- coding: utf-8 -*-

import pytest
from pmdarima.compat.pytest import pytest_error_str
from pmdarima.preprocessing.endog import LogEndogTransformer


def test_value_error_on_check():
    trans = LogEndogTransformer()  # could be anything, just need an instance
    with pytest.raises(ValueError) as ve:
        trans._check_y_X(None, None)
    assert 'non-None' in pytest_error_str(ve)



# -*- coding: utf-8 -*-

from pmdarima.datasets._base import load_date_example
from pmdarima.preprocessing.exog import DateFeaturizer
from pmdarima.compat.pytest import pytest_error_str

from numpy.testing import assert_array_equal
import pytest

y, X = load_date_example()


def test_no_options_warns():
    feat = DateFeaturizer(column_name="date",
                          with_day_of_month=False,
                          with_day_of_week=False)

    with pytest.warns(UserWarning) as w:
        y_prime, X_prime = feat.fit_transform(y, X)

    assert w is not None
    assert_array_equal(y, y_prime)
    assert X.equals(X_prime)


def test_illegal_column_fails():
    X_prime = X.copy()
    X_prime["date2"] = X_prime["date"].astype(str)

    feat = DateFeaturizer(column_name="date2")
    with pytest.raises(ValueError) as ve:
        feat.fit_transform(y, X_prime)

    assert "pd.Timestamp type" in pytest_error_str(ve)


def test_missing_column_fails():
    feat = DateFeaturizer(column_name="date2")
    with pytest.raises(ValueError) as ve:
        feat.fit_transform(y, X)

    assert "must exist" in pytest_error_str(ve)


def test_numpy_array_fails():
    feat = DateFeaturizer(column_name="date")
    with pytest.raises(TypeError) as te:
        feat.fit_transform(y, X.values)

    assert "X must be" in pytest_error_str(te)


def _dummy_assertions(X_prime):
    # they are dummies, so they should sum to 1 along the row axis
    dummies = X_prime[[n for n in X_prime.columns if 'WEEKDAY' in n]]
    assert (dummies.values.sum(axis=1) == 1).all()


def _ordinal_assertions(X_prime):
    # it's the day of the month, so they should all be > 0
    series = X_prime["DATE-DAY-OF-MONTH"]
    assert (series.values.ravel() > 0).all()


def test_all_true():
    feat = DateFeaturizer(column_name="date",
                          with_day_of_month=True,
                          with_day_of_week=True)

    y_prime, X_prime = feat.fit_transform(y, X)

    assert_array_equal(y, y_prime)
    assert y is not y_prime

    # there should be 8 columns in the X_prime (7 for days of the week, 1 for
    # ordinal)
    assert X_prime.shape[1] == 8

    _dummy_assertions(X_prime)
    _ordinal_assertions(X_prime)

    # date column should not be there anymore
    assert "date" not in X_prime.columns.tolist()


def test_dummy_only():
    feat = DateFeaturizer(column_name="date",
                          prefix="DATE",
                          with_day_of_month=False,
                          with_day_of_week=True)

    y_prime, X_prime = feat.fit_transform(y, X)

    assert_array_equal(y, y_prime)
    assert y is not y_prime

    # there should be 7 columns in the X_prime (7 for days of the week)
    assert X_prime.shape[1] == 7

    _dummy_assertions(X_prime)

    # show ordinal col not here
    assert "DATE-DAY-OF-MONTH" not in X_prime.columns.tolist()

    # date column should not be there anymore
    assert "date" not in X_prime.columns.tolist()


def test_ordinal_only():
    feat = DateFeaturizer(column_name="date",
                          prefix="DATE",
                          with_day_of_month=True,
                          with_day_of_week=False)

    y_prime, X_prime = feat.fit_transform(y, X)

    assert_array_equal(y, y_prime)
    assert y is not y_prime

    # there should be 1 column in the X_prime df
    assert X_prime.shape[1] == 1

    _ordinal_assertions(X_prime)

    # show ordinal col not here
    assert not [n for n in X_prime.columns.tolist() if "WEEKDAY" in n]

    # date column should not be there anymore
    assert "date" not in X_prime.columns.tolist()



# -*- coding: utf-8 -*-

import pandas as pd

from ..compat.numpy import DTYPE
from ._base import fetch_from_web_or_disk

__all__ = [
    'load_gasoline'
]

url = 'http://alkaline-ml.com/datasets/gasoline.csv'


def load_gasoline(as_series=False, dtype=DTYPE):
    """Weekly US finished motor gasoline products

    A weekly time series of US finished motor gasoline products supplied (in
    thousands of barrels per day) from February 1991 to May 2005.

    Parameters
    ----------
    as_series : bool, optional (default=False)
        Whether to return a Pandas series. If True, the index will be set to
        the observed years/months. If False, will return a 1d numpy array.

    dtype : type, optional (default=np.float64)
        The type to return for the array. Default is np.float64, which is used
        throughout the package as the default type.

    Notes
    -----
    The seasonal periodicity of this example is rather difficult, since it's
    not an integer. To be exact, the periodicity is ``365.25 / 7``
    (~=52.1785714285714). To fit the best possible model to this data, you'll
    need to explore using exogenous features

    See Also
    --------
    :class:`pmdarima.preprocessing.exog.FourierFeaturizer`

    Examples
    --------
    >>> from pmdarima.datasets import load_gasoline
    >>> load_gasoline()
    array([6621. , 6433. , 6582. , ..., 9024. , 9175. , 9269. ])

    >>> load_gasoline(True).head()
    0    6621.0
    1    6433.0
    2    6582.0
    3    7224.0
    4    6875.0
    dtype: float64

    References
    ----------
    .. [1] http://www.eia.gov/dnav/pet/hist/LeafHandler.ashx?n=PET&s=wgfupus2&f=W
    .. [2] https://robjhyndman.com/hyndsight/forecasting-weekly-data/

    Returns
    -------
    rslt : array-like, shape=(n_samples,)
        The gasoline dataset. There are 745 examples.
    """  # noqa
    rslt = fetch_from_web_or_disk(url, 'gasoline', cache=True).astype(dtype)
    if not as_series:
        return rslt

    return pd.Series(rslt)



"""
Utilities useful during the build -- adapted from sklearn.
"""
# author: Andy Mueller, Gael Varoquaux
# license: BSD

from numpy.distutils.system_info import get_info
from distutils.version import LooseVersion

import os
import contextlib

from .pre_build_helpers import basic_check_build

DEFAULT_ROOT = 'pmdarima'
CYTHON_MIN_VERSION = '0.28.5'  # 28 since 3.5 uses 28, not 29


def get_blas_info():
    def atlas_not_found(blas_info_):
        def_macros = blas_info.get('define_macros', [])
        for x in def_macros:
            if x[0] == "NO_ATLAS_INFO":
                # if x[1] != 1 we should have lapack
                # how do we do that now?
                return True
            if x[0] == "ATLAS_INFO":
                if "None" in x[1]:
                    # this one turned up on FreeBSD
                    return True
        return False

    blas_info = get_info('blas_opt', 0)
    if (not blas_info) or atlas_not_found(blas_info):
        cblas_libs = ['cblas']
        blas_info.pop('libraries', None)
    else:
        cblas_libs = blas_info.pop('libraries', [])

    return cblas_libs, blas_info


def _check_cython_version():
    message = 'Please install Cython with a version >= {0} in order ' \
              'to build a pmdarima distribution from source.' \
              .format(CYTHON_MIN_VERSION)
    try:
        import Cython
    except ModuleNotFoundError:
        # Re-raise with more informative error message instead:
        raise ModuleNotFoundError(message)

    if LooseVersion(Cython.__version__) < CYTHON_MIN_VERSION:
        message += (' The current version of Cython is {} installed in {}.'
                    .format(Cython.__version__, Cython.__path__))
        raise ValueError(message)


def cythonize_extensions(top_path, config):
    """Check that a recent Cython is available and cythonize extensions"""
    _check_cython_version()
    from Cython.Build import cythonize

    # Fast fail before cythonization if compiler fails compiling basic test
    # code even without OpenMP
    basic_check_build()

    # check simple compilation with OpenMP. If it fails scikit-learn will be
    # built without OpenMP and the test test_openmp_supported in the test suite
    # will fail.
    # `check_openmp_support` compiles a small test program to see if the
    # compilers are properly configured to build with OpenMP. This is expensive
    # and we only want to call this function once.
    # The result of this check is cached as a private attribute on the sklearn
    # module (only at build-time) to be used twice:
    # - First to set the value of SKLEARN_OPENMP_PARALLELISM_ENABLED, the
    #   cython build-time variable passed to the cythonize() call.
    # - Then in the build_ext subclass defined in the top-level setup.py file
    #   to actually build the compiled extensions with OpenMP flags if needed.

    n_jobs = 1
    with contextlib.suppress(ImportError):
        import joblib
        if LooseVersion(joblib.__version__) > LooseVersion("0.13.0"):
            # earlier joblib versions don't account for CPU affinity
            # constraints, and may over-estimate the number of available
            # CPU particularly in CI (cf loky#114)
            n_jobs = joblib.cpu_count()

    config.ext_modules = cythonize(
        config.ext_modules,
        nthreads=n_jobs,
        compiler_directives={'language_level': 3})


def gen_from_templates(templates, top_path):
    """Generate cython files from a list of templates"""
    # Lazy import because cython is not a runtime dependency.
    from Cython import Tempita

    for template in templates:
        outfile = template.replace('.tp', '')

        # if the template is not updated, no need to output the cython file
        if not (os.path.exists(outfile) and
                os.stat(template).st_mtime < os.stat(outfile).st_mtime):

            with open(template, "r") as f:
                tmpl = f.read()

            tmpl_ = Tempita.sub(tmpl)

            with open(outfile, "w") as f:
                f.write(tmpl_)



"""
Helpers to check build environment before actual build of pmdarima

Adapted from: https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/_build_utils/pre_build_helpers.py
"""  # noqa

import os
import sys
import glob
import tempfile
import textwrap
import subprocess

from distutils.sysconfig import customize_compiler
from numpy.distutils.ccompiler import new_compiler


def compile_test_program(code, extra_preargs=[], extra_postargs=[]):
    """Check that some C code can be compiled and run"""
    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    # extra_(pre/post)args can be a callable to make it possible to get its
    # value from the compiler
    if callable(extra_preargs):
        extra_preargs = extra_preargs(ccompiler)
    if callable(extra_postargs):
        extra_postargs = extra_postargs(ccompiler)

    start_dir = os.path.abspath('.')

    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            os.chdir(tmp_dir)

            # Write test program
            with open('test_program.c', 'w') as f:
                f.write(code)

            os.mkdir('objects')

            # Compile, test program
            ccompiler.compile(['test_program.c'], output_dir='objects',
                              extra_postargs=extra_postargs)

            # Link test program
            objects = glob.glob(
                os.path.join('objects', '*' + ccompiler.obj_extension))
            ccompiler.link_executable(objects, 'test_program',
                                      extra_preargs=extra_preargs,
                                      extra_postargs=extra_postargs)

            if "PYTHON_CROSSENV" not in os.environ:
                # Run test program if not cross compiling
                # will raise a CalledProcessError if return code was non-zero
                output = subprocess.check_output('./test_program')
                output = output.decode(sys.stdout.encoding or 'utf-8').splitlines() # noqa
            else:
                output = []
        except Exception:
            raise
        finally:
            os.chdir(start_dir)

    return output


def basic_check_build():
    """Check basic compilation and linking of C code"""
    code = textwrap.dedent(
        """\
        #include <stdio.h>
        int main(void) {
        return 0;
        }
        """)
    compile_test_program(code)



# -*- coding: utf-8 -*-

from numpy.testing import assert_array_almost_equal
from sklearn.base import clone

from pmdarima.preprocessing import LogEndogTransformer
from pmdarima.preprocessing import BoxCoxEndogTransformer


def test_same():
    y = [1, 2, 3]
    trans = BoxCoxEndogTransformer(lmbda=0)
    log_trans = LogEndogTransformer()
    y_t, _ = trans.fit_transform(y)
    log_y_t, _ = log_trans.fit_transform(y)
    assert_array_almost_equal(log_y_t, y_t)


def test_invertible():
    y = [1, 2, 3]
    trans = LogEndogTransformer()
    y_t, _ = trans.fit_transform(y)
    y_prime, _ = trans.inverse_transform(y_t)
    assert_array_almost_equal(y, y_prime)


def test_log_clone_issue_407():
    # https://github.com/alkaline-ml/pmdarima/issues/407
    log = LogEndogTransformer(lmbda=10)
    res, _ = log.fit_transform([0, 10])

    # we swap lmbda2 and lmbda internally
    assert log.lmbda2 == 10
    assert log.lmbda == 0

    log2 = clone(log)
    assert log2.lmbda2 == 10
    assert log2.lmbda == 0
    res2, _ = log2.fit_transform([0, 10])

    assert_array_almost_equal(res, res2)



# -*- coding: utf-8 -*-
#
# Author: Taylor Smith <taylor.smith@alkaline-ml.com>
#
# Common ARIMA functions

from sklearn.utils.validation import column_or_1d
import numpy as np

import warnings

from .. import context_managers as ctx
from ..utils import get_callable
from ..utils.array import diff, check_endog
from ..compat.numpy import DTYPE
from . import stationarity as statest_lib
from . import seasonality as seatest_lib

__all__ = [
    'is_constant',
    'ndiffs',
    'nsdiffs'
]

VALID_TESTS = {
    'kpss': statest_lib.KPSSTest,
    'adf': statest_lib.ADFTest,
    'pp': statest_lib.PPTest
}

VALID_STESTS = {
    'ocsb': seatest_lib.OCSBTest,
    'ch': seatest_lib.CHTest
}


def is_constant(x):
    """Test ``x`` for constancy.

    Determine whether a vector is composed of all of the same elements
    and nothing else.

    Parameters
    ----------
    x : array-like, shape=(n_samples,)
        The time series vector.

    Examples
    --------
    >>> import numpy as np
    >>> x = np.array([1, 2, 3])
    >>> y = np.ones(3)
    >>> [is_constant(x), is_constant(y)]
    [False, True]
    """
    x = column_or_1d(x)  # type: np.ndarray
    return (x == x[0]).all()


def nsdiffs(x, m, max_D=2, test='ocsb', **kwargs):
    """Estimate the seasonal differencing term, ``D``.

    Perform a test of seasonality for different levels of ``D`` to
    estimate the number of seasonal differences required to make a given time
    series stationary. Will select the maximum value of ``D`` for which
    the time series is judged seasonally stationary by the statistical test.

    Parameters
    ----------
    x : array-like, shape=(n_samples, [n_features])
        The array to difference.

    m : int
        The number of seasonal periods (i.e., frequency of the
        time series)

    max_D : int, optional (default=2)
        Maximum number of seasonal differences allowed. Must
        be a positive integer. The estimated value of ``D`` will not
        exceed ``max_D``.

    test : str, optional (default='ocsb')
        Type of unit root test of seasonality to use in order
        to detect seasonal periodicity. Valid tests include ("ocsb", "ch").
        Note that the CHTest is very slow for large data.

    Returns
    -------
    D : int
        The estimated seasonal differencing term. This is the maximum value
        of ``D`` such that ``D <= max_D`` and the time series is judged
        seasonally stationary. If the time series is constant, will return 0.
    """
    if max_D <= 0:
        raise ValueError('max_D must be a positive integer')

    # get the test - this validates m internally
    testfunc = get_callable(test, VALID_STESTS)(m, **kwargs)\
        .estimate_seasonal_differencing_term
    x = check_endog(x, dtype=DTYPE, copy=False)

    if is_constant(x):
        return 0

    D = 0
    dodiff = testfunc(x)
    while dodiff == 1 and D < max_D:
        D += 1
        x = diff(x, lag=m)

        if is_constant(x):
            return D

        # Issue 351: if the differenced array is now shorter than the seasonal
        # periodicity, we need to bail out now.
        if len(x) < m:
            warnings.warn("Appropriate D value may not have been reached; "
                          "length of seasonally-differenced array (%i) is "
                          "shorter than m (%i). Using D=%i"
                          % (len(x), m, D))
            return D

        dodiff = testfunc(x)

    return D


def ndiffs(x, alpha=0.05, test='kpss', max_d=2, **kwargs):
    """Estimate ARIMA differencing term, ``d``.

    Perform a test of stationarity for different levels of ``d`` to
    estimate the number of differences required to make a given time
    series stationary. Will select the maximum value of ``d`` for which
    the time series is judged stationary by the statistical test.

    Parameters
    ----------
    x : array-like, shape=(n_samples, [n_features])
        The array (time series) to difference.

    alpha : float, optional (default=0.05)
        Level of the test. This is the value above below which the P-value
        will be deemed significant.

    test : str, optional (default='kpss')
        Type of unit root test of stationarity to use in order to
        test the stationarity of the time-series. One of ('kpss', 'adf', 'pp')

    max_d : int, optional (default=2)
        Maximum number of non-seasonal differences allowed. Must
        be a positive integer. The estimated value of ``d`` will not
        exceed ``max_d``.

    Returns
    -------
    d : int
        The estimated differencing term. This is the maximum value of ``d``
        such that ``d <= max_d`` and the time series is judged stationary.
        If the time series is constant, will return 0.

    References
    ----------
    .. [1] R's auto_arima ndiffs function
           https://github.com/robjhyndman/forecast/blob/19b0711e554524bf6435b7524517715658c07699/R/arima.R#L132  # noqa: E501
    """
    if max_d <= 0:
        raise ValueError('max_d must be a positive integer')

    # get the test
    testfunc = get_callable(test, VALID_TESTS)(alpha, **kwargs).should_diff
    x = check_endog(x, dtype=DTYPE, copy=False)

    # base case, if constant return 0
    d = 0
    if is_constant(x):
        return d

    with ctx.except_and_reraise(
            np.linalg.LinAlgError,
            raise_err=ValueError,
            raise_msg="Encountered exception in stationarity test (%r). "
                      "This can occur in seasonal settings when a large "
                      "enough `m` coupled with a large enough `D` difference "
                      "the training array into too few samples for OLS "
                      "(input contains %i samples). Try fitting on a larger "
                      "training size" % (test, len(x)),
    ):
        # get initial diff
        pval, dodiff = testfunc(x)

        # if initially NaN, return 0
        if np.isnan(pval):
            return 0  # (d is zero, but this is more explicit to the reader)

        # Begin loop.
        while dodiff and d < max_d:
            d += 1

            # do differencing
            x = diff(x)
            if is_constant(x):
                return d

            # get new result
            pval, dodiff = testfunc(x)

            # if it's NaN now, take the last non-null one
            if np.isnan(pval):
                return d - 1

    # when d >= max_d
    return d




from pmdarima.utils.metaestimators import if_has_delegate


class _IfHasDelegateTester(object):
    def __init__(self):
        pass

    def fit(self):
        self.a_ = None
        return self

    @if_has_delegate('a_')
    def predict(self):
        return True

    @if_has_delegate(['b_', 'a_'])
    def predict2(self):
        return True


def test_single_delegate():
    # show it passes for a "fit"
    assert _IfHasDelegateTester().fit().predict()
    assert not hasattr(_IfHasDelegateTester(), 'predict')


def test_multiple_delegates():
    assert _IfHasDelegateTester().fit().predict2()
    assert not hasattr(_IfHasDelegateTester(), 'predict2')
